/* Example openCL saxpy code for ECE574 -- Spring 2023 */
/* By Vince Weaver <vincent.weaver@maine.edu> */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <stdint.h>
#include <errno.h>
#include <math.h>

/* Target OpenCL version 2.20 */
#define CL_TARGET_OPENCL_VERSION 220

#include <CL/cl.h>

const char *saxpy_kernel = "\n"
	"__kernel\n"
	"void opencl_saxpy(\n"
	"   const unsigned int n,\n"
	"   const float a,\n"
	"   __global float *x,\n"
	"   __global float *y) {\n"
	"\n"
	"   int i = get_global_id(0);\n"
	"\n"
	"   if (i < n) {\n"
	"      y[i] = a * x[i] + y[i];\n"
	"   }\n"
	"}\n"
	"\n";

/* Note: OpenCL does not include a strerror() function */
/* We provide this one to help out */
static const char *cl_getErrorString(cl_int error) {
	switch(error){
		/* run-time and JIT compiler errors */
		case CL_SUCCESS:
			return "CL_SUCCESS";				// 0
		case CL_DEVICE_NOT_FOUND:
			return "CL_DEVICE_NOT_FOUND";			// -1
		case CL_DEVICE_NOT_AVAILABLE:
			return "CL_DEVICE_NOT_AVAILABLE";		// -2
		case CL_COMPILER_NOT_AVAILABLE:
			return "CL_COMPILER_NOT_AVAILABLE";		// -3
		case CL_MEM_OBJECT_ALLOCATION_FAILURE:
			return "CL_MEM_OBJECT_ALLOCATION_FAILURE";	// -4
		case CL_OUT_OF_RESOURCES:
			return "CL_OUT_OF_RESOURCES";			// -5
		case -6:
			return "CL_OUT_OF_HOST_MEMORY";			// -6
		case -7:
			return "CL_PROFILING_INFO_NOT_AVAILABLE";	// -7
		case -8:
			return "CL_MEM_COPY_OVERLAP";			// -8
		case -9:
			return "CL_IMAGE_FORMAT_MISMATCH";		// -9
		case -10:
			return "CL_IMAGE_FORMAT_NOT_SUPPORTED";		// -10
		case -11:
			return "CL_BUILD_PROGRAM_FAILURE";		// -11
		case -12:
			return "CL_MAP_FAILURE";			// -12
		case -13:
			return "CL_MISALIGNED_SUB_BUFFER_OFFSET";	// -13
		case -14:
			return "CL_EXEC_STATUS_ERROR_FOR_EVENTS_IN_WAIT_LIST";	// -14
		case -15:
			return "CL_COMPILE_PROGRAM_FAILURE";		// -15
		case -16:
			return "CL_LINKER_NOT_AVAILABLE";		// -16
		case -17:
			return "CL_LINK_PROGRAM_FAILURE";		// -17
		case -18:
			return "CL_DEVICE_PARTITION_FAILED";		// -18
		case -19:
			return "CL_KERNEL_ARG_INFO_NOT_AVAILABLE";	// -19

		/* compile-time errors */
		case -30:
			return "CL_INVALID_VALUE";
		case -31:
			return "CL_INVALID_DEVICE_TYPE";
		case -32:
			return "CL_INVALID_PLATFORM";
		case -33:
			return "CL_INVALID_DEVICE";
		case -34:
			return "CL_INVALID_CONTEXT";
		case -35:
			return "CL_INVALID_QUEUE_PROPERTIES";
		case -36:
			return "CL_INVALID_COMMAND_QUEUE";
		case -37:
			return "CL_INVALID_HOST_PTR";
		case -38:
			return "CL_INVALID_MEM_OBJECT";
		case -39:
			return "CL_INVALID_IMAGE_FORMAT_DESCRIPTOR";
		case -40:
			return "CL_INVALID_IMAGE_SIZE";
		case -41:
			return "CL_INVALID_SAMPLER";
		case -42:
			return "CL_INVALID_BINARY";
		case -43:
			return "CL_INVALID_BUILD_OPTIONS";
		case -44:
			return "CL_INVALID_PROGRAM";
		case -45:
			return "CL_INVALID_PROGRAM_EXECUTABLE";
		case -46:
			return "CL_INVALID_KERNEL_NAME";
		case -47:
			return "CL_INVALID_KERNEL_DEFINITION";
		case -48:
			return "CL_INVALID_KERNEL";
		case -49:
			return "CL_INVALID_ARG_INDEX";
		case -50:
			return "CL_INVALID_ARG_VALUE";
		case -51:
			return "CL_INVALID_ARG_SIZE";
		case -52:
			return "CL_INVALID_KERNEL_ARGS";
		case -53:
			return "CL_INVALID_WORK_DIMENSION";
		case -54:
			return "CL_INVALID_WORK_GROUP_SIZE";
		case -55:
			return "CL_INVALID_WORK_ITEM_SIZE";
		case -56:
			return "CL_INVALID_GLOBAL_OFFSET";
		case -57:
			return "CL_INVALID_EVENT_WAIT_LIST";
		case -58:
			return "CL_INVALID_EVENT";
		case -59:
			return "CL_INVALID_OPERATION";
		case -60:
			return "CL_INVALID_GL_OBJECT";
		case -61:
			return "CL_INVALID_BUFFER_SIZE";
		case -62:
			return "CL_INVALID_MIP_LEVEL";
		case -63:
			return "CL_INVALID_GLOBAL_WORK_SIZE";
		case -64:
			return "CL_INVALID_PROPERTY";
		case -65:
			return "CL_INVALID_IMAGE_DESCRIPTOR";
		case -66:
			return "CL_INVALID_COMPILER_OPTIONS";
		case -67:
			return "CL_INVALID_LINKER_OPTIONS";
		case -68:
			return "CL_INVALID_DEVICE_PARTITION_COUNT";

		/* extension errors */
		case -1000:
			return "CL_INVALID_GL_SHAREGROUP_REFERENCE_KHR";// -1000
		case -1001:
			return "CL_PLATFORM_NOT_FOUND_KHR";		// -1001
		case -1002:
			return "CL_INVALID_D3D10_DEVICE_KHR";		// -1002
		case -1003:
			return "CL_INVALID_D3D10_RESOURCE_KHR";		// -1003
		case -1004:
			return "CL_D3D10_RESOURCE_ALREADY_ACQUIRED_KHR";// -1004
		case -1005:
			return "CL_D3D10_RESOURCE_NOT_ACQUIRED_KHR";	// -1005
		default:
			return "Unknown OpenCL error";
	}
}


#define PLATFORM_INFO_LENGTH 1024

int main(int argc, char **argv) {

	/* OpenCL defines */
#define MAX_PLATFORMS	5
	cl_int err;
	cl_device_id device_id;
	cl_platform_id *platform_ids;
	cl_uint num_platforms;
	size_t returned_size = 0;
	struct platform_info_t {
		cl_char platform_name[PLATFORM_INFO_LENGTH];
		cl_char platform_prof[PLATFORM_INFO_LENGTH];
		cl_char platform_vers[PLATFORM_INFO_LENGTH];
		cl_char platform_exts[PLATFORM_INFO_LENGTH];
	} *platform_info;
	cl_uint num_devices;
	cl_device_id *device_ids;
	cl_context context;
	cl_command_queue commands;
	cl_program program_saxpy;
	cl_kernel kernel_saxpy;
	size_t local;
	size_t global;
	int i;
	int which_platform=0;


	/****************************/
	/* Check command line usage */
	/****************************/
	if (argc<1) {
		fprintf(stderr,"Usage: %s\n",argv[0]);
		return -1;
	}
	if (argc>1) {
		which_platform=atoi(argv[1]);
	}


	/*********************/
	/* Initialize OpenCL */
	/*********************/

	/* Query number of platforms */

	err = clGetPlatformIDs(0,NULL,&num_platforms);
	if (err != CL_SUCCESS) {
		fprintf(stderr,"Error: Failed to get number of platforms! %s\n",
			cl_getErrorString(err));
		return EXIT_FAILURE;
	}

	/* allocate space for platform ids */
	platform_ids=calloc(num_platforms,sizeof(cl_platform_id));
	if (platform_ids==NULL) {
		fprintf(stderr,"Error allocating memory for platform ids!\n");
		return EXIT_FAILURE;
	}

	/* Get Platforms */
	/* Arguments are: size of the platform array */
	/*                pointer to platform array  */
	/*                number of platforms available */
	err = clGetPlatformIDs(num_platforms,platform_ids,&num_platforms);
	if (err != CL_SUCCESS) {
		fprintf(stderr,"Error: Failed to get a platform id! %s\n",
			cl_getErrorString(err));
		return EXIT_FAILURE;
	}

	printf("\nOpenCL platform information, found %d platforms:\n\n",
		num_platforms);

	/* Get platform info */

	/* allocate space for platform info */
	platform_info=calloc(num_platforms,sizeof(struct platform_info_t));
	if (platform_info==NULL) {
		fprintf(stderr,"Error allocating memory for platform info!\n");
		return EXIT_FAILURE;
	}

	for(i=0;i<num_platforms;i++) {
		err  = clGetPlatformInfo(platform_ids[i], CL_PLATFORM_NAME,
			PLATFORM_INFO_LENGTH, platform_info[i].platform_name,
			&returned_size);
		if (err != CL_SUCCESS) {
			printf("Error: Failed to get platform info! %s\n",
				cl_getErrorString(err));
			return EXIT_FAILURE;
		}

		err = clGetPlatformInfo(platform_ids[i], CL_PLATFORM_VERSION,
			PLATFORM_INFO_LENGTH, platform_info[i].platform_vers,
			&returned_size);
		if (err != CL_SUCCESS) {
			printf("Error: Failed to get platform info! %s\n",
				cl_getErrorString(err));
			return EXIT_FAILURE;
		}

		err = clGetPlatformInfo(platform_ids[i], CL_PLATFORM_PROFILE,
			PLATFORM_INFO_LENGTH, platform_info[i].platform_prof,
			&returned_size);
		if (err != CL_SUCCESS) {
			printf("Error: Failed to get platform info! %s\n",
				cl_getErrorString(err));
			return EXIT_FAILURE;
		}

		err = clGetPlatformInfo(platform_ids[i], CL_PLATFORM_EXTENSIONS,
			PLATFORM_INFO_LENGTH, platform_info[i].platform_exts,
			&returned_size);
		if (err != CL_SUCCESS) {
			printf("Error: Failed to get platform info! %s\n",
				cl_getErrorString(err));
			return EXIT_FAILURE;
		}

		printf("Platform %d:\n",i);
		printf("\tPlatform name:       %s\n", (char *)
			platform_info[i].platform_name);
		printf("\tPlatform version:    %s\n", (char *)
			platform_info[i].platform_vers);
		printf("\tPlatform profile:    %s\n", (char *)
			platform_info[i].platform_prof);
		printf("\tPlatform extensions: %s\n",
			(platform_info[i].platform_exts[0] != '\0') ?
			(char *)platform_info[i].platform_exts : "NONE");
		printf("\n");

	}

	/* Print which one we are using */

	printf("-------------------------------------\n");
	printf("Using platform %d (%s) for calculations.\n\n",
		which_platform,platform_info[which_platform].platform_name);


	/****************************/
	/* Initializing the Devices */
	/****************************/

	/* detect number of devices */

	err = clGetDeviceIDs(platform_ids[which_platform],
			CL_DEVICE_TYPE_ALL,
			0, NULL, &num_devices);

	/* allocate space for device ids */
	device_ids=calloc(num_devices,sizeof(cl_device_id));
	if (device_ids==NULL) {
		fprintf(stderr,"Error allocating memory for device ids!\n");
		return EXIT_FAILURE;
	}

	err = clGetDeviceIDs(platform_ids[which_platform],
			CL_DEVICE_TYPE_ALL,
			num_devices, device_ids, &num_devices);

	if (err != CL_SUCCESS) {
		printf("Failed to collect device list on this platform! %s\n",
			cl_getErrorString(err));
		return EXIT_FAILURE;
	}

	printf("Found %d compute device for this platform.\n",num_devices);


	/* Print device name */

	unsigned char device_name[1024];

	err=clGetDeviceInfo(device_ids[0],CL_DEVICE_NAME,1024,&device_name,0);
	printf("Device is %s\n",device_name);

	/* Create a compute context for the requested device */
	/* We assume device 0 */

	device_id=device_ids[0];

	context = clCreateContext(NULL, 1, &device_id, NULL, NULL, &err);

	if (!context) {
		printf("Error: Failed to create a compute context! %s\n",
			cl_getErrorString(err));
		return EXIT_FAILURE;
	}

	/* Create a command queue */

	commands = clCreateCommandQueueWithProperties(
			context, device_id, NULL, & err);

	if (!commands) {
		printf("Error: Failed to create a command queue! %s\n",
			cl_getErrorString(err));
		return EXIT_FAILURE;
	}


	/******************************/
	/* Create the saxpy kernel    */
	/******************************/

	/* arguments are: the context
			how many strings,
			pointer to the strings
			string length array (NULL if they are NUL terminated)
			error return */
	program_saxpy = clCreateProgramWithSource(context, 1,
			(const char **) &saxpy_kernel, NULL, &err);

	if ((!program_saxpy) || (err!=CL_SUCCESS)) {
		printf("Error: Failed to create compute program! %s\n",
						cl_getErrorString(err));
		return EXIT_FAILURE;
	}

	/* Build the program */

	/* arguments are:	program,
				num_devices (devices in dev_list)
				dev_list (NULL means build for all)
				options: string to pass to compiler
				pfn_notify: callback when compiling done
				user_data: used with pfn_notify */
	err = clBuildProgram(program_saxpy, 0, NULL, NULL, NULL, NULL);

	/* Print errors from the kernel compile */
	if (err != CL_SUCCESS) {
    		size_t len;
    		char buffer[2048];
   		printf("Error: Failed to build program executable! %s\n",
						cl_getErrorString(err));
		clGetProgramBuildInfo(program_saxpy, device_id,
			CL_PROGRAM_BUILD_LOG, sizeof(buffer), buffer, &len);
		printf("%s\n", buffer);
		exit(1);
	}


	/* Create kernel from the compiled code */

	kernel_saxpy = clCreateKernel(program_saxpy,
					"opencl_saxpy", &err);

	if (!kernel_saxpy || err != CL_SUCCESS) {
		printf("Error: Failed to create compute kernel %s!\n",
						cl_getErrorString(err));
		exit(1);
	}

	/*************************************/
	/* Allocate space on device          */
	/*************************************/

	int N=1000*1000*8;

	cl_mem dev_x;
	cl_mem dev_y;

	float *x,*y;
	float a;

	a=5.0;

	/* Allocate vectors on CPU */
        x=(float *)malloc(N*sizeof(float));
        y=(float *)malloc(N*sizeof(float));

	/* Initialize the host vectors */
        for(i=0;i<N;i++) {
                x[i]=(float)i;
                y[i]=(float)(10.0*i);
        }


	/* Allocate vectors on device */

	/* Input data, read-only on device */
	dev_x = clCreateBuffer(context, CL_MEM_READ_ONLY,
				sizeof(float) * N, NULL, NULL);

	/* Input/output, read/write */
	dev_y = clCreateBuffer(context, CL_MEM_READ_WRITE,
				sizeof(float) * N, NULL, NULL);


	err  = clEnqueueWriteBuffer(commands, dev_x, CL_TRUE, 0,
			sizeof(float) * N, x, 0, NULL, NULL);
	if (err != CL_SUCCESS) {
		printf("Error copying! %s\n", cl_getErrorString(err));
		exit(1);
	}

	err  = clEnqueueWriteBuffer(commands, dev_y, CL_TRUE, 0,
			sizeof(float) * N, y, 0, NULL, NULL);
	if (err != CL_SUCCESS) {
		printf("Error copying! %s\n", cl_getErrorString(err));
		exit(1);
	}

	/* Set the kernel arguments */
	err  = 0;
	err |= clSetKernelArg(kernel_saxpy, 0, sizeof(unsigned int), &N);
	err |= clSetKernelArg(kernel_saxpy, 1, sizeof(float), &a);
	err |= clSetKernelArg(kernel_saxpy, 2, sizeof(cl_mem), &dev_x);
	err |= clSetKernelArg(kernel_saxpy, 3, sizeof(cl_mem), &dev_y);

	if (err != CL_SUCCESS) {
		printf("Error: Failed to set kernel arguments! sobel_x\n");
		exit(1);
	}

	/* Get the size of a thread group on this platform */
	err = clGetKernelWorkGroupInfo(kernel_saxpy, device_id,
		CL_KERNEL_WORK_GROUP_SIZE, sizeof(local), &local, NULL);

	if (err != CL_SUCCESS) {
		printf("Error: Failed to retrieve kernel work "
				"group info! %d\n", err);
		exit(1);
	}

	/* Set the size of our problem (number of pixels to calculate) */

	global = N;
	printf("Problem size is %d, dividing up in thread blocks of %ld\n",
		 (int)global,local);


	/* arguments are:
			command queue,
			kernel to run,
			number of dimensions of thread/block/grid
			work offset (future: currently always NULL)
			work size: number of elements
			local size: thread split up (NULL means auto)
			num in wait list:
			wait_list: list of events that must finish before run
			events: returns event object for this request
	*/
	err = clEnqueueNDRangeKernel(commands, kernel_saxpy,
			1, NULL, &global, NULL, 0, NULL, NULL);

	if (err) {
		printf("Error: Failed to execute kernel (Convolve x)! %s\n",
					cl_getErrorString(err));
		return EXIT_FAILURE;
	}

	/* Finish and synchronize (is this necessary?) */
	clFinish(commands);



	/* Copy the results back to host */
	err = clEnqueueReadBuffer( commands, dev_y, CL_TRUE, 0,
		sizeof(float) * N, y, 0, NULL, NULL );

	if (err != CL_SUCCESS) {
		printf("Error: Failed to read output array! %d\n", err);
		exit(1);
	}

	/* results */
        i=100;
        printf("y[%d]=%f, y[%d]=%f\n",i,y[i],N-1,y[N-1]);

        /* y[i]=a*x[i]+y[i] */
        /* 0: a=5, x=0, y=0  ::::::: y=0 */
        /* 1: a=5, x=1, y=10 ::::::: y=15 */
        /* 2: a=5, x=2, y=20 ::::::: y=30 */
        /* 3: a=5, x=3, y=30 ::::::: y=45 */
        /* 4: a=5, x=4, y=40 ::::::: y=60 */
        /* ... */
        /* 100: a=5, x=100, y=1000 y=1500 */

	/* TODO: free everything */


	return 0;
}
